﻿using System;
using System.ComponentModel;
using System.IO;
using System.Text;
using System.Timers;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using CH.Regatron.HPPS.EnergyStorage.BatSimulation;
using CH.Regatron.HPPS.EnergyStorage.BatSimulation.StateMachine;
using Microsoft.Win32;
using TopConAPIDotNet4WPFExample.Commands;

namespace TopConAPIDotNet4WPFExample.BatterySimulation
{
  public class BatterySimulationViewModel : INotifyPropertyChanged
  {
    public delegate void SimulationInfoDelegate(bool running);
    public event SimulationInfoDelegate OnSimulationInfo;

    #region //---- Members ----

    private readonly Brush _scriptRunningOnColor = new SolidColorBrush(Colors.LawnGreen);
    private readonly Brush _scriptRunningOffColor = new SolidColorBrush(Colors.LightGray);

    private readonly Brush _simulationActivOnColor = new SolidColorBrush(Colors.LawnGreen);
    private readonly Brush _simulationActivOffColor = new SolidColorBrush(Colors.LightGray);


    private ICommand _getBatteryConfigurationPathCommand;
    private ICommand _loadScriptCommand;
    private readonly CH.Regatron.HPPS.EnergyStorage.BatSimulation.BatterySimulation _batterySimulation;
    private string _selectedPath;
    private string _script;
    private readonly Timer _valueUpdateTimer;

    private string _simulationState;
    private string _stateOfCharge;
    private string _errorMessage;

    private bool _readyToRun;
    private string _scriptPath;

    private ICommand _runSimulationCommand;
    private ICommand _pauseSimulationCommand;
    private ICommand _resumeSimulationCommand;
    private ICommand _stopSimulationCommand;

    #endregion

    #region //---- Properties ----
    
    public string SimulationVoltageOutput { get; set; }

    public string SimulationActualCurrent { get; set; }

    public string SimulationChargeCount { get; set; }

    public string SimulationTotalCapacity { get; set; }

    public string SimulationStateOfCharge { get; set; }

    public string SimulationCutoffInformation { get; set; }

    public Visibility ChargeImageVisibility { get; set; }

    public Visibility DischargeImageVisibility { get; set; }

    public Visibility IdleImageVisibility { get; set; }

    public double ProgressBarStateOfChargeValue { get; set; }

    public string SimulationState
    {
      get { return _simulationState; }
      set
      {
        _simulationState = value;
        OnPropertyChanged(nameof(SimulationState));
      }
    }

    public string StateOfCharge
    {
      get { return _stateOfCharge; }
      set
      {
        _stateOfCharge = value;
        OnPropertyChanged(nameof(StateOfCharge));
      }
    }

    public string ErrorMessage
    {
      get { return _errorMessage; }

      set
      {
        _errorMessage = value;
        OnPropertyChanged(nameof(ErrorMessage));
      }
    }

    public Brush ScriptRunningColor
    {
      get
      {
        if(ScriptIsRunning)
        {
          return _scriptRunningOnColor;
        }

        return _scriptRunningOffColor;
      }
    }

    private bool ScriptIsRunning { get; set; }

    public Brush SimulationActivColor
    {
      get
      {
        if(SimulationActiv)
        {
          return _simulationActivOnColor;
        }

        return _simulationActivOffColor;
      }
    }

    public bool SimulationActiv { get; set; }

    public string SelectedPath
    {
      get { return _selectedPath; }
      set
      {
        _selectedPath = value;


        if(string.IsNullOrEmpty(_selectedPath) == false)
        {
          try
          {
            _batterySimulation.LoadBatteryConfiguration(_selectedPath);

            //set StateOfCharge to actual battery value after loading configuration
            _batterySimulation.GetBatteryActualValues(out ActualBatteryValues actualBatteryValues);
            StateOfCharge = actualBatteryValues.StateOfCharge.ToString("F3");
          }
          catch
          {
            _selectedPath = _batterySimulation.GetLastError();
          }
        }

        OnPropertyChanged(nameof(SelectedPath));
        ReadyToRun = true;
      }
    }

    public bool ReadyToRun
    {
      get { return _readyToRun; }
      set
      {
        _readyToRun = _batterySimulation.ReadyToRun;
        OnPropertyChanged(nameof(ReadyToRun));
      }
    }

    public bool IsBatSimAvailable
    {
      get
      {
        if (_batterySimulation != null)
        {
          return _batterySimulation.IsBatSimAvailable();
        }
        else
        {
          return false;
        }
      }
    }

    public string Script
    {
      get { return _script; }
      set
      {
        try
        {
          if(_batterySimulation.LoadCustomScript(value))
          {
            ScriptPath = value;
            _script = File.ReadAllText(value);
          }
        }
        catch
        {
          _script = _batterySimulation.GetLastError();
        }
        OnPropertyChanged(nameof(Script));
        ReadyToRun = true;
      }
    }

    public string ScriptPath
    {
      get { return _scriptPath; }
      set
      {
        _scriptPath = value;
        OnPropertyChanged(nameof(ScriptPath));
      }
    }

    #endregion

    public BatterySimulationViewModel(CH.Regatron.HPPS.EnergyStorage.BatSimulation.BatterySimulation batterySimulation)
    {
      _batterySimulation = batterySimulation;
      _batterySimulation.OnSimulationStateChanged += SimulationStateChangedDelegate;
      _batterySimulation.OnConnectionChanged += ConnectionChangedHandler;
      ReadyToRun = false;
      _valueUpdateTimer = new Timer(1000);
      _valueUpdateTimer.Elapsed += ValueUpdateTimerOnElapsed;
      IdleImageVisibility = Visibility.Visible;
      OnPropertyChanged(nameof(IdleImageVisibility));
      OnPropertyChanged(nameof(IsBatSimAvailable));
    }

    private void SetErrorMessage(Exception ex)
    {
      BatterySimulationStateException sbse = ex as BatterySimulationStateException;
      if(sbse != null)
      {
        StringBuilder sb = new StringBuilder();
        sb.AppendLine($"Actual state = {sbse.ActualSimulationState}");
        sb.AppendLine($"Desired state = {sbse.DesiredSimulationState}");
        sb.Append(sbse.Message);
        ErrorMessage = sb.ToString();
      }
      else
      {
        ErrorMessage = ex.Message;
      }
    }

    public void UdateConnectionStateProperties()
    {
      OnPropertyChanged(nameof(IsBatSimAvailable));
    }

    public void TopConDisconnected()
    {
      UdateConnectionStateProperties();
    }

    #region //---- Callbacks ----

    private void SimulationStateChangedDelegate(BatterySimulationState simulationState)
    {
      if (simulationState == BatterySimulationState.Idle ||
          simulationState == BatterySimulationState.Stopped)
      {
        _valueUpdateTimer.Stop() ;
      }
      else
      {
        _valueUpdateTimer.Start();
      }

      SimulationState = simulationState.ToString();
      OnPropertyChanged(nameof(ReadyToRun));
      OnPropertyChanged(nameof(IsBatSimAvailable));
    }

    private void ValueUpdateTimerOnElapsed(object sender, ElapsedEventArgs elapsedEventArgs)
    {
      ActualBatteryValues actualBatteryValues;

      if(_batterySimulation.GetBatteryActualValues(out actualBatteryValues))
      {
        SimulationVoltageOutput = actualBatteryValues.VoltageOutput.ToString("F3");
        OnPropertyChanged(nameof(SimulationVoltageOutput));

        SimulationActualCurrent = actualBatteryValues.ActualCurrent.ToString("F3");
        OnPropertyChanged(nameof(SimulationActualCurrent));

        SimulationChargeCount = actualBatteryValues.ChargeCount.ToString("F3");
        OnPropertyChanged(nameof(SimulationChargeCount));

        SimulationCutoffInformation = actualBatteryValues.CutoffInformation;
        OnPropertyChanged(nameof(SimulationCutoffInformation));

        SimulationTotalCapacity = actualBatteryValues.MaximumCapacity.ToString("F3");
        OnPropertyChanged(nameof(SimulationTotalCapacity));

        ProgressBarStateOfChargeValue = actualBatteryValues.StateOfCharge;
        OnPropertyChanged(nameof(ProgressBarStateOfChargeValue));

        SimulationStateOfCharge = actualBatteryValues.StateOfCharge.ToString("F3");
        OnPropertyChanged(nameof(SimulationStateOfCharge));

        if(actualBatteryValues.ChargingState == ActualBatteryValues.BatteryChargingState.Charge)
        {
          ChargeImageVisibility = Visibility.Visible;
          DischargeImageVisibility = Visibility.Hidden;
          IdleImageVisibility = Visibility.Hidden;
        }
        else if(actualBatteryValues.ChargingState == ActualBatteryValues.BatteryChargingState.Discharge)
        {
          ChargeImageVisibility = Visibility.Hidden;
          DischargeImageVisibility = Visibility.Visible;
          IdleImageVisibility = Visibility.Hidden;
        }
        else if(actualBatteryValues.ChargingState == ActualBatteryValues.BatteryChargingState.Idle)
        {
          ChargeImageVisibility = Visibility.Hidden;
          DischargeImageVisibility = Visibility.Hidden;
          IdleImageVisibility = Visibility.Visible;
        }

        ScriptIsRunning = _batterySimulation.ScriptIsRunning;
        OnPropertyChanged(nameof(ScriptRunningColor));

        SimulationActiv = _batterySimulation.IsSimulationActiv();
        OnPropertyChanged(nameof(SimulationActivColor));

        OnPropertyChanged(nameof(ChargeImageVisibility));
        OnPropertyChanged(nameof(DischargeImageVisibility));
        OnPropertyChanged(nameof(IdleImageVisibility));
      }
    }

    private void ConnectionChangedHandler(bool connected, int port)
    {
      UdateConnectionStateProperties();
    }

    #endregion

    #region //---- Commands ----

    #region //-- InitCommand --

    public ICommand GetBatteryConfigurationPathCommand
    {
      get
      {
        if(_getBatteryConfigurationPathCommand == null)
        {
          _getBatteryConfigurationPathCommand = new GenericCommand(GetBatteryConfigurationPath);
        }
        return _getBatteryConfigurationPathCommand;
      }
    }

    private void GetBatteryConfigurationPath(object obj)
    {
      var dialog = new OpenFileDialog
      {
        InitialDirectory = String.Format($"{Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData)}\\Regatron\\ESSControl\\configuration"),
        RestoreDirectory = true
      };

      dialog.ShowDialog();

      SelectedPath = dialog.FileName;
    }

    #endregion

    #region //-- LoadScriptCommand --

    public ICommand LoadScriptCommand
    {
      get
      {
        if(_loadScriptCommand == null)
        {
          _loadScriptCommand = new GenericCommand(LoadScriptHandler);
        }
        return _loadScriptCommand;
      }
    }

    private void LoadScriptHandler(object obj)
    {
      var dialog = new OpenFileDialog
      {
        InitialDirectory = String.Format($"{Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData)}\\Regatron\\ESSControl"),
        RestoreDirectory = true
      };

      dialog.ShowDialog();

      Script = dialog.FileName;
    }

    #endregion

    #region //-- Run Simulation --

    public ICommand RunSimulationCommand
    {
      get
      {
        if(_runSimulationCommand == null)
        {
          _runSimulationCommand = new GenericCommand(RunSimulationHandler);
        }
        return _runSimulationCommand;
      }
    }

    private void RunSimulationHandler(object obj)
    {
      ErrorMessage = string.Empty;
      try
      {
        double soc;
        if (Double.TryParse(_stateOfCharge, out soc))
        {
          ErrorMessage = string.Empty;
          try
          {
            _batterySimulation.SetStateOfCharge(soc);
          }
          catch (Exception ex)
          {
            SetErrorMessage(ex);
          }
        }

        _batterySimulation.StartSimulation();
        if(OnSimulationInfo != null)
        {
          OnSimulationInfo(true);
        }

      }
      catch(Exception ex)
      {
        SetErrorMessage(ex);
      }
    }

    #endregion

    #region //-- Pause Simulation --

    public ICommand PauseSimulationCommand
    {
      get
      {
        if(_pauseSimulationCommand == null)
        {
          _pauseSimulationCommand = new GenericCommand(PauseSimulationHandler);
        }
        return _pauseSimulationCommand;
      }
    }

    private void PauseSimulationHandler(object obj)
    {
      ErrorMessage = string.Empty;
      try
      {
        _batterySimulation.PauseSimulation();
      }
      catch(Exception ex)
      {
        SetErrorMessage(ex);
      }
    }

    #endregion

    #region //-- Resume Simulation --

    public ICommand ResumeSimulationCommand
    {
      get
      {
        if(_resumeSimulationCommand == null)
        {
          _resumeSimulationCommand = new GenericCommand(ResumeSimulationHandler);
        }
        return _resumeSimulationCommand;
      }
    }

    private void ResumeSimulationHandler(object obj)
    {
      ErrorMessage = string.Empty;
      
      try
      {
        _batterySimulation.ResumeSimulation();
      }
      catch(Exception ex)
      {
        SetErrorMessage(ex);
      }
    }

    #endregion

    #region //-- Stop Simulation --

    public ICommand StopSimulationCommand
    {
      get
      {
        if(_stopSimulationCommand == null)
        {
          _stopSimulationCommand = new GenericCommand(StopSimulationHandler);
        }
        return _stopSimulationCommand;
      }
    }

    private void StopSimulationHandler(object obj)
    {
      ErrorMessage = string.Empty;
      try
      {
        bool success = _batterySimulation.StopSimulation();
        if(success)
        {
          ScriptIsRunning = _batterySimulation.ScriptIsRunning;
          SimulationActiv = _batterySimulation.IsSimulationActiv();
          OnPropertyChanged(nameof(ScriptRunningColor));
          OnPropertyChanged(nameof(SimulationActivColor));

          //Update Icon State
          ChargeImageVisibility = Visibility.Hidden;
          DischargeImageVisibility = Visibility.Hidden;
          IdleImageVisibility = Visibility.Visible;
          OnPropertyChanged(nameof(ChargeImageVisibility));
          OnPropertyChanged(nameof(DischargeImageVisibility));
          OnPropertyChanged(nameof(IdleImageVisibility));

          if(OnSimulationInfo != null)
          {
            OnSimulationInfo(false);
          }
        }
      }
      catch(Exception ex)
      {
        SetErrorMessage(ex);
      }
    }

    #endregion

    #endregion
    
    #region //---- IPropertyChanged Interface ----

    public event PropertyChangedEventHandler PropertyChanged;

    private void OnPropertyChanged(string propertyName)
    {
      PropertyChangedEventHandler handler = PropertyChanged;
      if(handler != null) handler(this, new PropertyChangedEventArgs(propertyName));
    }

    #endregion
  }
}