﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Reflection;
using System.Threading;
using System.Threading.Tasks;
using System.Timers;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using TopConAPIDotNet4WPFExample.BatterySimulation;
using TopConAPIDotNet4WPFExample.Commands;
using TopConAPIDotNet4WPFExample.SasSimulation;
using BatterySimulationView = TopConAPIDotNet4WPFExample.BatterySimulation.BatterySimulationView;
using DEV = CH.Regatron.HPPS.Device;

using Timer = System.Timers.Timer;

namespace TopConAPIDotNet4WPFExample
{
  public sealed class MainWindowViewModel:INotifyPropertyChanged
  {
    #region //---- Members ----
    
    private ICommand _initCommand;
    private ICommand _connectCommand;
    private ICommand _disConnectCommand;
    private ICommand _powerOnCommand;
    private ICommand _powerOffCommand;
    private ICommand _clearErrorCommand;
    private uint _portNr;
    private DEV.TopCon _topCon;
    private readonly Brush _powerOnColor = new SolidColorBrush(Colors.LawnGreen);
    private readonly Brush _powerOffColor = new SolidColorBrush(Colors.LightGray);

    private readonly Brush _errorColorNoError = new SolidColorBrush(Colors.LawnGreen);
    private readonly Brush _errorColorWarning = new SolidColorBrush(Colors.Orange);
    private readonly Brush _errorColorError = new SolidColorBrush(Colors.Red);

    private float _actualVoltage;
    private float _actualCurrent;
    private float _actualPower;

    private CancellationTokenSource _tokenSource;
    CancellationToken _cancelToken;
    private const int UpdateIntervallMs = 200;
    private Task _valueUpdateTask;

    private BatterySimulationViewModel _batterySimulationViewModel;
    private readonly BatterySimulationView _batterySimulationView;

    private SasSimulationViewModel _sasSimulationViewModel;
    private readonly SasSimulationView _sasSimulationView;

    #endregion

    #region //---- Properties ----

    public uint PortNr
    {
      get { return _portNr; }
      set
      {
        _portNr = value;
        OnPropertyChanged("PortNr");
      }
    }

    public bool IsConnected
    {
      get
      {
        bool retVal = false;
        if (_topCon != null)
        {
          try
          { 
            int comPortNr = _topCon.GetConnectedRs232ComPortNumber();
            retVal = comPortNr > 0;
          }
          catch (Exception e)
          {

          }
        }
        return retVal;
      }
    }

    public bool IsPowerOn
    {
      get
      {
        bool retVal = false;
        if(IsConnected)
        {
          try
          {
            retVal = _topCon.IsPowerOn();
          }
          catch(Exception e)
          {

          }
        }
        return retVal;
      }
    }

    public Brush PowerColor
    {
      get
      {
        if(IsPowerOn)
        {
          return _powerOnColor;
        }
        else
        {
          return _powerOffColor;
        }
      }
    }

    public Brush ErrorColor
    {
      get
      {
        if(IsConnected)
        {
          if (_topCon.IsInWarningState())
          {
            return _errorColorWarning;
          }
          else if (_topCon.IsInErrorState())
          {
            return _errorColorError;
          }
          return _errorColorNoError;
        }
        else
        {
          return _powerOffColor;
        }
      }
    }

    public string ActualVoltage
    {
      get { return String.Format("{0:F2} V", _actualVoltage); }
    }

    public string ActualCurrent
    {
      get { return String.Format("{0:F2} A", _actualCurrent); }
    }

    public string ActualPower
    {
      get { return String.Format("{0:F2} W", _actualPower); }
    }

    public bool IsSasEnabled
    {
      get
      {
        if(SasSimulationViewModel != null)
        {
          return SasSimulationViewModel.IsSasSimAvailable;
        }
        else
        {
          return false;
        }
      }
    }

    public bool IsBatSimEnabled
    {
      get
      {
        if(BatterySimulationViewModel != null)
        {
          return BatterySimulationViewModel.IsBatSimAvailable;
        }
        else
        {
          return false;
        }
      }
    }

    public BatterySimulationViewModel BatterySimulationViewModel
    {
      get
      {
        return _batterySimulationViewModel;
      }
      set
      {
        _batterySimulationViewModel = value;
        _batterySimulationView.DataContext = _batterySimulationViewModel;
        OnPropertyChanged(nameof(BatterySimulationViewModel));
      }
    }
    
    public SasSimulationViewModel SasSimulationViewModel
    {
      get
      {
        return _sasSimulationViewModel;
      }
      set
      {
        _sasSimulationViewModel = value;
        _sasSimulationView.DataContext = _sasSimulationViewModel;
        OnPropertyChanged(nameof(SasSimulationViewModel));
      }
    }

    #endregion

    /// <summary>
    /// Constructor
    /// </summary>
    public MainWindowViewModel(BatterySimulationView batterySimulationView, SasSimulationView sasSimulationView)
    {
      _batterySimulationView = batterySimulationView;
      _sasSimulationView = sasSimulationView;
      PortNr = 1;
      UpdateConnectionProperties();
    }

    private void UpdateConnectionProperties()
    {
      OnPropertyChanged(nameof(IsConnected));
      OnPropertyChanged(nameof(PowerColor));
      OnPropertyChanged(nameof(ErrorColor));
      OnPropertyChanged(nameof(IsSasEnabled));
      OnPropertyChanged(nameof(IsBatSimEnabled));

      if (SasSimulationViewModel != null)
      {
        SasSimulationViewModel.UdateConnectionStateProperties();
      }

      if (BatterySimulationViewModel != null)
      {
        BatterySimulationViewModel.UdateConnectionStateProperties();
      }
    }

    #region //---- Value Updater ----

    private void ValueUpdateWorker(CancellationToken cancelToken)
    {
      try
      {
        while(!_cancelToken.IsCancellationRequested)
        {
          if(IsConnected)
          {
            _actualVoltage = (float)_topCon.GetActualVoltage();
            _actualCurrent = (float)_topCon.GetActualCurrent();
            _actualPower = (float)_topCon.GetActualPower();

            OnPropertyChanged(nameof(ActualVoltage));
            OnPropertyChanged(nameof(ActualCurrent));
            OnPropertyChanged(nameof(ActualPower));
            OnPropertyChanged(nameof(ErrorColor));
          }
          Task.Delay(UpdateIntervallMs, _cancelToken).Wait(_cancelToken);
        }

        Debug.WriteLine("ValueUpdater ended");
      }
      catch(Exception e)
      {
        Console.WriteLine(e);
      }
    }

    private bool IsValueUpdateTaskActiv()
    {
      return _valueUpdateTask?.Status == TaskStatus.Running;
    }

    private void StartValueUpdateTask()
    {
      if(_tokenSource == null)
      {
        _tokenSource = new CancellationTokenSource();
        _cancelToken = _tokenSource.Token;
      }

      if(IsConnected && !IsValueUpdateTaskActiv())
      {
        _valueUpdateTask = Task.Factory.StartNew(() => ValueUpdateWorker(_cancelToken), _cancelToken, TaskCreationOptions.LongRunning, TaskScheduler.Default);
      }
    }

    private void StopValueUpdateTask()
    {
      if(IsValueUpdateTaskActiv())
      {
        _tokenSource.Cancel();

        try
        {
          _valueUpdateTask.Wait();

          _tokenSource = null;
          _valueUpdateTask = null;
        }
        catch(Exception e)
        {
          if(e?.InnerException is TaskCanceledException || e is OperationCanceledException)
          {
            //dont do anything
          }
          else
          {
            throw;
          }
        }
      }
    }

    #endregion

    #region //---- Callbacks ----

    private void BatterySimulationInfoCallback(bool running)
    {
      if(running)
      {
        SasSimulationViewModel.StopLiveViewer();
      }
    }

    #endregion
    
    #region //---- Commands ----

    #region //-- InitCommand --

    public ICommand InitCommand
    {
      get
      {
        if(_initCommand == null)
        {
          _initCommand = new GenericCommand(InitTopConDevice);
        }
        return _initCommand;
      }
    }

    private void InitTopConDevice(object obj)
    {
      Mouse.OverrideCursor = Cursors.Wait;

      try
      {
        _topCon = new CH.Regatron.HPPS.Device.TopCon();
        string initInfo;
        if (_topCon != null)
        {
          initInfo = "Initialization successful";
        }
        else
        {
          initInfo = "Initialization FAILED";
        }
        InitInfo = initInfo;
        OnPropertyChanged(nameof(InitInfo));
      }
      catch (Exception e)
      {
        MessageBox.Show(String.Format("Initialization FAILED. Error: {0}", e), "Init", MessageBoxButton.OK, MessageBoxImage.Error); 
      }
      finally
      {
        Mouse.OverrideCursor = null;
      }
    }

    public string InitInfo { get; set; }

    #endregion

    #region //-- ConnectCommand --

    public ICommand ConnectCommand
    {
      get
      {
        if(_connectCommand == null)
        {
          _connectCommand = new GenericCommand(Connect, canConnect => !IsConnected);
        }
        return _connectCommand;
      }
    }

    private void Connect(object obj)
    {
      Mouse.OverrideCursor = Cursors.Wait;

      try
      {
        if (_topCon != null)
        {
          _topCon.Connect((int)PortNr);

          StartValueUpdateTask();

          if (BatterySimulationViewModel == null)
          {
            var batterySimulation = _topCon.GetBatterySimulationObject();
            if (batterySimulation == null)
            {
              MessageBox.Show("BatSim option is not installed in the TC device", "Device option not installed", MessageBoxButton.OK, MessageBoxImage.Information);
            }
            else
            {
              BatterySimulationViewModel = new BatterySimulationViewModel(batterySimulation);
              BatterySimulationViewModel.OnSimulationInfo += BatterySimulationInfoCallback;
            }
          }
          
          if (SasSimulationViewModel == null)
          {
            SasSimulationViewModel = new SasSimulationViewModel(_topCon);

            if (!SasSimulationViewModel.IsSasSimAvailable)
            {
              MessageBox.Show("SAS option is not installed in the TC device", "Device option not installed", MessageBoxButton.OK, MessageBoxImage.Information);
            }
          }
        }
        else
        {
          MessageBox.Show("Cannot connect until initialization succeeded", "Connect", MessageBoxButton.OK, MessageBoxImage.Error);
        }
      }
      catch(Exception e)
      {
        MessageBox.Show(String.Format("Connection FAILED. \n\nError: \n{0}", e), "Init", MessageBoxButton.OK, MessageBoxImage.Error);
      }
      finally
      {
        Mouse.OverrideCursor = null;
        UpdateConnectionProperties();
      }
    }

    #endregion

    #region //-- DisConnectCommand --

    public ICommand DisconnectCommand
    {
      get
      {
        if(_disConnectCommand == null)
        {
          _disConnectCommand = new GenericCommand(Disconnect, canDisconnect => IsConnected);
        }
        return _disConnectCommand;
      }
    }

    private void Disconnect(object obj)
    {
      Mouse.OverrideCursor = Cursors.Wait;

      try
      {
        if(_topCon != null)
        {
          DoDisconnect();
        }
        else
        {
          MessageBox.Show("Cannot Disconnect", "Disconnect", MessageBoxButton.OK, MessageBoxImage.Error);
        }
      }
      catch(Exception e)
      {
        MessageBox.Show(String.Format("Disconnect FAILED. \n\nError: \n{0}", e), "Init", MessageBoxButton.OK, MessageBoxImage.Error);
      }
      finally
      {
        Mouse.OverrideCursor = null;
        UpdateConnectionProperties();
      }
    }

    internal void DoDisconnect()
    {
      try
      {
        if(_topCon != null)
        {
          if(SasSimulationViewModel != null)
          {
            SasSimulationViewModel.StopLiveViewer();
          }

          if(BatterySimulationViewModel != null && BatterySimulationViewModel.SimulationActiv)
          {
            BatterySimulationViewModel.StopSimulationCommand.Execute(null);
          }

          StopValueUpdateTask();

          _topCon.Disconnect();
        }
      }
      finally
      {
        UpdateConnectionProperties();
      }
      
    }

    #endregion

    #region //-- PowerOnCommand --

    public ICommand PowerOnCommand
    {
      get
      {
        if(_powerOnCommand == null)
        {
          _powerOnCommand = new GenericCommand(DoPowerOn, canPowerOn => IsConnected && !IsPowerOn);
        }
        return _powerOnCommand;
      }
    }

    private void DoPowerOn(object obj)
    {
      Mouse.OverrideCursor = Cursors.Wait;

      try
      {
        if(IsConnected)
        {
          _topCon.SetPowerOn();
        }
        else
        {
          MessageBox.Show("Could net set Power ON. Connect device first.", "Power", MessageBoxButton.OK, MessageBoxImage.Error);
        }
      }
      catch(Exception e)
      {
        MessageBox.Show(String.Format("Power ON FAILED. \n\nError: \n{0}", e), "Init", MessageBoxButton.OK, MessageBoxImage.Error);
      }
      finally
      {
        Mouse.OverrideCursor = null;
        OnPropertyChanged(nameof(IsPowerOn));
        OnPropertyChanged(nameof(PowerColor));
      }
    }

    #endregion

    #region //-- PowerOffCommand --

    public ICommand PowerOffCommand
    {
      get
      {
        if(_powerOffCommand == null)
        {
          _powerOffCommand = new GenericCommand(DoPowerOff, canPowerOff => IsConnected && IsPowerOn);
        }
        return _powerOffCommand;
      }
    }

    private void DoPowerOff(object obj)
    {
      Mouse.OverrideCursor = Cursors.Wait;

      try
      {
        if(IsConnected)
        {
          _topCon.SetPowerOff();
        }
        else
        {
          MessageBox.Show("Could not set Power to OFF. Connect first to device.", "Power", MessageBoxButton.OK, MessageBoxImage.Error);
        }
      }
      catch(Exception e)
      {
        MessageBox.Show(String.Format("Power OFF FAILED. \n\nError: \n{0}", e), "Init", MessageBoxButton.OK, MessageBoxImage.Error);
      }
      finally
      {
        Mouse.OverrideCursor = null;
        OnPropertyChanged(nameof(IsPowerOn));
        OnPropertyChanged(nameof(PowerColor));
        OnPropertyChanged(nameof(ErrorColor));
      }
    }

    #endregion

    #region //-- ClearErrorCommand --

    public ICommand ClearErrorCommand
    {
      get
      {
        if(_clearErrorCommand == null)
        {
          _clearErrorCommand = new GenericCommand(ClearError, canPowerOn => IsConnected);
        }
        return _clearErrorCommand;
      }
    }

    private void ClearError(object obj)
    {
      Mouse.OverrideCursor = Cursors.Wait;

      try
      {
        if(IsConnected)
        {
          _topCon.ClearError();
        }
        else
        {
          MessageBox.Show("Could not clear Errors. Connect device first.", "Power", MessageBoxButton.OK, MessageBoxImage.Error);
        }
      }
      catch(Exception e)
      {
        MessageBox.Show(String.Format("Clear Error FAILED. \n\nError: \n{0}", e), "Init", MessageBoxButton.OK, MessageBoxImage.Error);
      }
      finally
      {
        Mouse.OverrideCursor = null;
        OnPropertyChanged(nameof(ErrorColor));
      }
    }

    #endregion

    #endregion

    #region //---- IPropertyChanged Interface ----

    public event PropertyChangedEventHandler PropertyChanged;

    private void OnPropertyChanged(string propertyName)
    {
      PropertyChangedEventHandler handler = PropertyChanged;
      if (handler != null) handler(this, new PropertyChangedEventArgs(propertyName));
    }

    #endregion
  }
}
